from pprint import pformat
import pandas as pd
from sklearn.ensemble import IsolationForest
from sklearn.preprocessing import MinMaxScaler
from gordo.machine.dataset.datasets import pandas_filter_rows
import logging
from typing import Optional

logger = logging.getLogger(__name__)


class WrongFilterMethodType(TypeError):
    pass


class FilterPeriods:
    """Model class with methods for data pre-processing.
    Performs a series of algorithms that drops noisy data.

    Either a rolling median or an isolation forest algorithm is executed.
    Both provide drop periods in a dict-type element on the class object
    `object.drop_periods["iforest"]` and `object.drop_periods["median"]`,
    and data is filtered accordingly.

    Parameters:
    data: pandas.DataFrame
        Data frame containing already filtered data (global max/min + dropped known periods).
        Time consecutively is not required.
    granularity: str
        The bucket size for grouping all incoming time data (e.g. "10T").
        Available strings come from https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
    **kwargs:
        See below.

    Keyword arguments:
    filter_method: str
        Which method should be used for data cleaning, either "median" (default), "iforest" or "all" which returns
        results for both methods.
    iforest_smooth: bool
        If exponential weighted smoothing should be applied to data before isolation forest
        algorithm is run.
    """

    def __init__(
        self,
        granularity: Optional[str] = "10T",
        filter_method: Optional[str] = "median",
        window=144,
        n_iqr=5,
        iforest_smooth=False,
        contamination=0.03,
    ):
        self.granularity = granularity
        self.filter_method = filter_method
        if self.filter_method not in ["median", "iforest", "all"]:
            raise WrongFilterMethodType
        self._window = window
        self._n_iqr = n_iqr
        self._iforest_smooth = iforest_smooth
        self._contamination = contamination

    def filter_data(self, data):
        """Method for filtering data.
        Returns the filtered dataset, a dict containing the different periods that have been dropped arranged
        by filtering method and the actual predictions from the filter model.
        """
        predictions = {}
        if self.filter_method in ["median", "all"]:
            predictions["median"] = self._rolling_median(data)

        if self.filter_method in ["iforest", "all"]:
            self._train(data)
            predictions["iforest"] = self._predict(data)

        drop_periods = self._drop_periods(predictions)
        data = self._filter_data(data, drop_periods)
        return data, drop_periods, predictions

    def _init_model(self, data):
        """Return a new instance of the models."""
        self.isolationforest = IsolationForest(
            n_estimators=300,  # The number of base estimators in the ensemble.
            max_samples=min(
                1000, data.shape[0]
            ),  # Samples to draw from X to train each base estimator
            contamination=self._contamination,  # default = "auto"
            max_features=1.0,  # Features to draw from X to train each base estimator.
            bootstrap=False,
            n_jobs=-1,  # ``-1`` means using all processors
            random_state=42,
            verbose=0,
        )
        self.minmaxscaler = MinMaxScaler()
        logger.info("Initialized isolationforest: \n%s.", self.isolationforest)
        logger.info("Initialized minmaxscaler: \n%s.", self.minmaxscaler)

    def _train(self, data):
        """Train the model.
        Smooth if necessary.
        """
        data = data.copy()
        if self._iforest_smooth:
            data = data.ewm(halflife=6).mean()

        logger.info("Fitting model")
        self._init_model(data)
        self.model = self.isolationforest.fit(data)

        logger.info(
            "Created new isolationforest model:\n%s\nFitted on data of shape '%s'.",
            self.model,
            data.shape,
        )

    def _predict(self, data):
        """Make predictions.
        """
        logger.info("Calculating predictions for isolation forest")
        assert isinstance(data, pd.DataFrame)

        score = -self.model.decision_function(data)
        self.iforest_scores = self._describe(score)
        score = self.minmaxscaler.fit_transform(score.reshape(-1, 1)).squeeze()
        self.iforest_scores_transformed = self._describe(score)

        pred = self.model.predict(data)
        pred = pd.DataFrame({"pred": pred, "score": score, "timestamp": data.index})
        logger.info("Anomaly ratio: %s", list(pred).count(-1) / pred.shape[0])
        return pred

    def _rolling_median(self, data):
        """Function for filtering using a rolling median approach."""
        logger.info("Calculating predictions for rolling median")
        roll = data.rolling(self._window, center=True)
        r_md = roll.median()
        r_iqr = roll.quantile(0.75) - roll.quantile(0.25)
        high = r_md + self._n_iqr * r_iqr
        low = r_md - self._n_iqr * r_iqr
        mask = ((data < low) | (data > high)).any(1).astype("int") * -1
        pred = pd.DataFrame({"pred": mask})
        pred.index.name = "timestamp"
        pred = pred.reset_index()
        logger.info("Anomaly ratio: %s", list(pred).count(-1) / pred.shape[0])
        return pred

    def _drop_periods(self, predictions):
        """Create drop period list.

        Only keep anomaly flagged observations (-1), and create a time-diff between these.
        Time consecutive anomalies will have a time-diff equal to the specified
        granularity of the data, e.g. 10 minutes.
        Minimum two consecutive anomalies must be flagged for a drop period to be initiated.
        """
        logger.info("Creating list of drop period dicts")
        drop_periods = {}

        if self.filter_method == "all":
            pred_types = ["iforest", "median"]

        else:
            pred_types = [self.filter_method]

        for pred_type in pred_types:
            t = predictions[pred_type].query("pred == -1")[["timestamp"]]
            t["delta_t"] = (
                t["timestamp"]
                .diff()
                .fillna(pd.Timedelta(seconds=0))
                .astype("timedelta64[m]")
                .astype("int")
            )
            t = t.reset_index(drop=True)

            # get granularity in minutes
            granularity = pd.Timedelta(self.granularity).total_seconds() / 60

            start = []
            end = []
            for i in range(len(t)):
                # start conditions:
                # [i == 0 (start) OR
                # time delta from previous > granularity (gap)]
                if (i == 0) or (t["delta_t"][i] > granularity):
                    start.append(str(t["timestamp"][i]))
                # stop conditions:
                # [i + 1 == len(t) (end of data) OR
                # time delta to next > granularity (gap)]
                if (i + 1 == len(t)) or (t["delta_t"][i + 1] > granularity):
                    end.append(str(t["timestamp"][i]))
            drop_periods[pred_type] = pd.DataFrame(
                {"drop_start": start, "drop_end": end}
            ).to_dict("records")

        return drop_periods

    def _filter_data(self, data, drop_periods):
        """Drops periods defined previously from dataset."""
        row_filter = []
        for drop_method, p in drop_periods.items():
            for line in p:
                row_filter.append(
                    f"~('{line['drop_start']}' <= index <= '{line['drop_end']}')"
                )

        if row_filter:
            n_prior = len(data)
            data = pandas_filter_rows(df=data, filter_str=row_filter, buffer_size=0)
            logger.info(f"Dropped {n_prior - len(data)} rows")
            return data
        else:
            logger.info("No rows dropped")
            return data

    @staticmethod
    def _describe(score):
        """Format array-like for logging summary statistics."""
        return pformat(pd.Series(score).describe().round(3).to_dict())



import logging
import pandas as pd
import numpy as np
from typing import Union

logger = logging.getLogger(__name__)


def apply_buffer(mask: pd.Series, buffer_size: int = 0):
    """
    Take a mask (boolean series) where True indicates keeping a value, and False
    represents removing the value. This will 'expand' those indexes marked as `False`
    to the symmetrical bounds of ``buffer_size``

    Parameters
    ----------
    mask: pandas.core.Series
        Boolean pandas series
    buffer_size: int
        Size to buffer around ``False`` values

    Examples
    --------
    >>> import pandas as pd
    >>> series = pd.Series([True, True, False, True, True])
    >>> series = apply_buffer(series, buffer_size=1)
    >>> series
    0     True
    1    False
    2    False
    3    False
    4     True
    dtype: bool

    Returns
    -------
    None
    """
    if (not any(mask)) or (buffer_size == 0):
        return mask

    array = 1 - np.array(mask.to_numpy(), dtype=float)
    kernel = np.ones(buffer_size * 2 + 1, dtype=float)

    if len(kernel) > len(array):
        mask.values[:] = False
        return mask

    ans = np.convolve(a=array, v=kernel, mode="same")
    mask.values[:] = ans < 1
    return mask


def pandas_filter_rows(
    df: pd.DataFrame, filter_str: Union[str, list], buffer_size: int = 0
):
    """ Filter pandas data frame based on list or string of conditions.

    Note:
    pd.DataFrame.eval of a list returns a numpy.ndarray and is limited to 100 list items.
    The sparse evaluation with numexpr pd.DataFrame.eval of a combined string logic, can only consist of
    a maximum 32 (current dependency) or 242 logical parts (latest release) and returns a pd.Series
    Therefore, list elements are evaluated in batches of n=15 (to be safe) and evaluate iterative.

    Parameters
    ----------
    df: pandas.Dataframe
      Dataframe to filter rows from. Does not modify the parameter
    filter_str: str or list
      String representing the filter. Can be a boolean combination of conditions,
      where conditions are comparisons of column names and either other columns
      or numeric values. The rows matching the filter are kept.
      Column names with spaces must be quoted with backticks,
      names without spaces could be quoted with backticks or be unquoted.
      Example of legal filters are " `Tag A` > 5 " , " (`Tag B` > 1) | (`Tag C` > 4)"
      '(`Tag D` < 5) ', " (TagB > 5) "
      The parameter can also be a list, in which the items will be joined by logical " & ".
    buffer_size: int
      Area fore and aft of the application of ``fitler_str`` to also mark for removal.

    Returns
    -------
    pandas.Dataframe
        The dataframe containing only rows matching the filter

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> df = pd.DataFrame(list(np.ndindex((3,3))), columns=list('AB'))
    >>> df
       A  B
    0  0  0
    1  0  1
    2  0  2
    3  1  0
    4  1  1
    5  1  2
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, "`A`>1")
       A  B
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, "`A`> B")
       A  B
    3  1  0
    6  2  0
    7  2  1
    >>> pandas_filter_rows(df, "(`A`>1) | (`B`<1)")
       A  B
    0  0  0
    3  1  0
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, "(`A`>1) & (`B`<1)")
       A  B
    6  2  0
    >>> pandas_filter_rows(df, ["A>1", "B<1"])
       A  B
    6  2  0
    >>> pandas_filter_rows(df, ["A!=1", "B<3"])
       A  B
    0  0  0
    1  0  1
    2  0  2
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, ["A!=1", "B<3"], buffer_size=1)
       A  B
    0  0  0
    1  0  1
    7  2  1
    8  2  2
    """
    logger.info("Applying numerical filtering to data of shape %s", df.shape)

    if isinstance(filter_str, str):
        mask = df.eval(filter_str)

    if isinstance(filter_str, list):
        mask = []
        for filter_i in _batch(iterable=filter_str, n=15):
            mask.append(df.eval(" & ".join(filter_i)))

        mask = pd.concat(mask, axis=1).all(axis=1)

    if buffer_size != 0:
        mask = apply_buffer(mask, buffer_size=buffer_size)

    df = df[list(mask)]
    logger.info("Shape of data after numerical filtering: %s", df.shape)
    return df


def _batch(iterable, n: int):
    """Helper function for creating batches on list items"""
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]



# -*- coding: utf-8 -*-

import pytest
from gordo.machine.dataset.datasets import RandomDataset
from gordo.machine.dataset.filter_periods import FilterPeriods
from gordo.machine.dataset.sensor_tag import SensorTag


@pytest.fixture
def dataset():
    return RandomDataset(
        train_start_date="2017-01-01 00:00:00Z",
        train_end_date="2018-01-01 00:00:00Z",
        tag_list=[SensorTag("Tag 1", None)],
    )


def test_filter_periods_typerror(dataset):
    data, _ = dataset.get_data()
    assert data.shape == (9760, 1)
    with pytest.raises(TypeError):
        FilterPeriods(granularity="10T", filter_method="abc", n_iqr=1)


def test_filter_periods_median(dataset):
    data, _ = dataset.get_data()
    data_filtered, drop_periods, predictions = FilterPeriods(
        granularity="10T", filter_method="median", n_iqr=1
    ).filter_data(data)

    assert data.shape == (9063, 1)
    assert data["Tag 1"].mean() == 0.5113691034704841

    assert sum(predictions["median"]["pred"]) == -493
    assert len(drop_periods["median"]) == 44
    assert data_filtered.shape == (8570, 1)


def test_filter_periods_iforest(dataset):
    data, _ = dataset.get_data()
    data_filtered, drop_periods, predictions = FilterPeriods(
        granularity="10T", filter_method="iforest", iforest_smooth=False
    ).filter_data(data)

    assert data.shape == (12838, 1)
    assert data["Tag 1"].mean() == 0.5144733352386245

    assert sum(predictions["iforest"]["pred"]) == 12066
    assert len(drop_periods["iforest"]) == 61
    assert data_filtered.shape == (12452, 1)


def test_filter_periods_all(dataset):
    data, _ = dataset.get_data()
    data_filtered, drop_periods, predictions = FilterPeriods(
        granularity="10T", filter_method="all", n_iqr=1, iforest_smooth=False
    ).filter_data(data)

    assert data.shape == (8024, 1)
    assert data["Tag 1"].mean() == 0.500105748646813

    assert sum(predictions["median"]["pred"]) == -449
    assert sum(predictions["iforest"]["pred"]) == 7542
    assert len(drop_periods["median"]) == 39
    assert len(drop_periods["iforest"]) == 29
    assert data_filtered.shape == (7356, 1)


def test_filter_periods_iforest_smoothing(dataset):
    data, _ = dataset.get_data()
    data_filtered, drop_periods, predictions = FilterPeriods(
        granularity="10T", filter_method="iforest", iforest_smooth=True
    ).filter_data(data)

    assert data.shape == (9674, 1)
    assert data["Tag 1"].mean() == 0.5019862352609169

    assert sum(predictions["iforest"]["pred"]) == 8552
    assert len(drop_periods["iforest"]) == 41
    assert data_filtered.shape == (9113, 1)


def test_filter_periods_all_smoothing(dataset):
    data, _ = dataset.get_data()
    data_filtered, drop_periods, predictions = FilterPeriods(
        granularity="10T", filter_method="all", n_iqr=1, iforest_smooth=True
    ).filter_data(data)

    assert data.shape == (8595, 1)
    assert data["Tag 1"].mean() == 0.512856120233814

    assert sum(predictions["iforest"]["pred"]) == 7471
    assert len(drop_periods["median"]) == 39
    assert len(drop_periods["iforest"]) == 29
    assert data_filtered.shape == (7522, 1)
